import * as React from 'react';

export interface UseVideoControlsOptions {
  mediaControllerRef: React.Ref<HTMLVideoElement>;
  actionFormKey?: string;
}

export const useVideoControls = ({
  mediaControllerRef: forwardedRef,
  actionFormKey = 'action' as const,
}: UseVideoControlsOptions) => {
  const defaultRef = React.useRef<HTMLVideoElement>(null);
  const mediaControllerRef = forwardedRef ?? defaultRef;
  const seekRef = React.useRef<HTMLInputElement>(null);
  const volumeRef = React.useRef<HTMLInputElement>(null);
  const filenameRef = React.useRef<HTMLElement>(null);
  const [isPlaying, setIsPlaying] = React.useState(false);
  const [isSeeking, setIsSeeking] = React.useState(false);
  const [currentTimeDisplay, setCurrentTimeDisplay] = React.useState<number>();
  const [seekTimeDisplay, setSeekTimeDisplay] = React.useState<number>();
  const [durationDisplay, setDurationDisplay] = React.useState<number>();
  const [isSeekTimeCountingDown, setIsSeekTimeCountingDown] = React.useState(false);

  const refreshControls: React.ReactEventHandler<HTMLVideoElement> = (e) => {
    const { currentTarget: mediaController } = e;
    const { current: seek } = seekRef;
    if (!seek) {
      return;
    }

    setCurrentTimeDisplay(mediaController.currentTime);
    setDurationDisplay(mediaController.duration);
  };

  const togglePlayback = React.useCallback(() => {
    setIsPlaying((p) => !p);
  }, []);

  const startSeek: React.MouseEventHandler<HTMLInputElement> = React.useCallback(() => {
    setIsSeeking(true);
  }, []);

  const doSetSeek = (thisElement: HTMLInputElement, mediaController: HTMLVideoElement) => {
    mediaController.currentTime = thisElement.valueAsNumber;
  };

  const setSeek: React.ChangeEventHandler<HTMLInputElement> = React.useCallback((e) => {
    if (!(typeof mediaControllerRef === 'object' && mediaControllerRef !== null)) {
      return;
    }

    const { current: mediaController } = mediaControllerRef;
    if (!mediaController) {
      return;
    }

    const { currentTarget: thisElement } = e;
    setSeekTimeDisplay(thisElement.valueAsNumber);

    if (isSeeking) {
      return;
    }

    doSetSeek(thisElement, mediaController);
  }, [mediaControllerRef, isSeeking]);

  const endSeek: React.MouseEventHandler<HTMLInputElement> = React.useCallback((e) => {
    if (!(typeof mediaControllerRef === 'object' && mediaControllerRef !== null)) {
      return;
    }

    const { current: mediaController } = mediaControllerRef;

    if (!mediaController) {
      return;
    }

    const { currentTarget: thisElement } = e;
    setIsSeeking(false);
    doSetSeek(thisElement, mediaController);
  }, [mediaControllerRef]);

  const resetVideo: React.ReactEventHandler<HTMLVideoElement> = React.useCallback((e) => {
    const videoElement = e.currentTarget;
    setIsPlaying(false);
    videoElement.currentTime = 0;
  }, []);

  const updateSeekFromPlayback: React.ReactEventHandler<HTMLVideoElement> = React.useCallback((e) => {
    if (!seekRef.current) {
      return;
    }

    if (isSeeking) {
      return;
    }

    const { current: seek } = seekRef;

    if (!seek) {
      return;
    }

    const videoElement = e.currentTarget;
    const currentTime = videoElement.currentTime;
    setCurrentTimeDisplay(currentTime);
    seek.value = String(currentTime);
  }, [isSeeking]);

  const toggleSeekTimeCountMode = React.useCallback(() => {
    setIsSeekTimeCountingDown((b) => !b);
  }, []);

  const download = React.useCallback(() => {
    if (!(typeof mediaControllerRef === 'object' && mediaControllerRef?.current !== null)) {
      return;
    }

    if (!(typeof filenameRef === 'object' && filenameRef?.current !== null)) {
      return;
    }

    const downloadLink = window.document.createElement('a');
    downloadLink.download = filenameRef.current.textContent ?? 'image';
    downloadLink.href = mediaControllerRef.current.currentSrc;
    downloadLink.addEventListener('click', () => {
      downloadLink.remove();
    });
    downloadLink.click();
  }, [mediaControllerRef, filenameRef]);

  const actions = React.useMemo(() => ({
    togglePlayback,
    toggleSeekTimeCountMode,
    download,
  }), [togglePlayback, toggleSeekTimeCountMode, download]);

  const handleAction: React.FormEventHandler<HTMLFormElement> = React.useCallback((e) => {
    e.preventDefault();
    const formData = new FormData(e.currentTarget, (e.nativeEvent as unknown as { submitter: HTMLElement }).submitter);
    const actionName = formData.get(actionFormKey) as keyof typeof actions;
    const { [actionName]: actionFunction } = actions;
    actionFunction?.();
  }, [actions, actionFormKey]);

  const adjustVolume: React.ChangeEventHandler<HTMLInputElement> = React.useCallback((e) => {
    if (!(typeof mediaControllerRef === 'object' && mediaControllerRef !== null)) {
      return;
    }

    if (!mediaControllerRef.current) {
      return;
    }
    const { value } = e.currentTarget;
    mediaControllerRef.current.volume = Number(value);
  }, [mediaControllerRef]);

  React.useEffect(() => {
    if (!(typeof mediaControllerRef === 'object' && mediaControllerRef !== null)) {
      return;
    }

    if (!mediaControllerRef.current) {
      return;
    }

    if (isPlaying) {
      void mediaControllerRef.current.play();
      return
    }

    mediaControllerRef.current.pause();
  }, [isPlaying, mediaControllerRef]);

  React.useEffect(() => {
    if (!seekRef.current) {
      return;
    }

    const { current: seek } = seekRef;
    if (!seek) {
      return;
    }

    seek.value = String(currentTimeDisplay);
  }, [currentTimeDisplay]);

  React.useEffect(() => {
    if (!seekRef.current) {
      return;
    }

    const { current: seek } = seekRef;
    if (!seek) {
      return;
    }

    seek.max = String(durationDisplay);
  }, [durationDisplay]);

  React.useEffect(() => {
    if (!volumeRef.current) {
      return;
    }

    if (!(typeof mediaControllerRef === 'object' && mediaControllerRef !== null)) {
      return;
    }

    if (!mediaControllerRef.current) {
      return;
    }

    const { current: mediaController } = mediaControllerRef;
    const { current: volume } = volumeRef;
    volume.value = String(mediaController.volume);
  }, [mediaControllerRef]);

  return React.useMemo(() => ({
    seekRef,
    volumeRef,
    isPlaying,
    refreshControls,
    adjustVolume,
    resetVideo,
    startSeek,
    endSeek,
    setSeek,
    updateSeekFromPlayback,
    durationDisplay,
    currentTimeDisplay,
    seekTimeDisplay,
    isSeeking,
    isSeekTimeCountingDown,
    mediaControllerRef,
    handleAction,
    filenameRef,
  }), [
    isPlaying,
    isSeeking,
    adjustVolume,
    resetVideo,
    startSeek,
    endSeek,
    setSeek,
    updateSeekFromPlayback,
    durationDisplay,
    currentTimeDisplay,
    seekTimeDisplay,
    isSeekTimeCountingDown,
    mediaControllerRef,
    handleAction,
    filenameRef,
  ]);
};
